#!/usr/bin/env ruby
#    Rubyripper - A secure ripper for Linux/BSD/OSX
#    Copyright (C) 2007  Bouke Woudstra (rubyripperdev@gmail.com)
#
#    This file is part of Rubyripper. Rubyripper is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>

#ensure translation is working before actually installing
LOCALE=[ENV['PWD'] + "/locale", "/usr/share/locale"]
LOCALE.each{|dir| if File.directory?(dir) : ENV['GETTEXT_PATH'] = dir ; break end}

$rr_version = '0.4.4' #application wide setting

begin
  require 'gettext'
rescue LoadError
  puts "RubyGems is not found. Please install ruby-gettext."
end

def installed(filename) # a help function to check if an application is installed
	ENV['PATH'].split(':').each do |dir|
		if File.exist?(dir + '/' + filename) : return true end
	end
	if File.exist?(filename) : return true else return false end #it can also be in current working dir
end

def cdrom_drive #default values for cdrom drives under differenty os'es
	drive = 'Unknown!'
	system = RUBY_PLATFORM
	if system.include?('linux') || system.include?('bsd')
		drive = '/dev/cdrom'
	elsif system.include?('darwin')
		drive = '/dev/disk1'
	elsif system.include?('win') # no support for Windows as of yet, but keeping possibilities open for the future
		drive = 'cdaudio'
	end
	return drive
end

def get_example_filename_normal(basedir, layout) #separate function to make it faster
	filename = File.join(basedir, layout)
	filename = File.expand_path(filename)
	filename = "Example filename: #{filename}.ext"
	{'%a' => 'Judas Priest', '%b' => 'Sin After Sin', '%f' => 'codec', '%g' => 'Rock', '%y' => '1977', '%n' =>'01', '%t' => 'Sinner', '%i' =>'inputfile', '%o' => 'outputfile'}.each{|key, value| filename.gsub!(key,value)}
	return filename
end

def get_example_filename_various(basedir, layout) #separate function to make it faster
	filename = File.join(basedir, layout)
	filename = File.expand_path(filename)
	filename = "Example filename: #{filename}.ext"
	{'%a' => 'Various Artists', '%b' => 'TMF Rockzone', '%f' => 'codec', '%g' => "Rock", '%y' => '1999', '%n' => '01', '%va' => 'Kid Rock', '%t' => 'Cowboy'}.each{|key, value| filename.gsub!(key,value)}
	return filename
end

def get_filename(settings, codec = 'mp3', track = 1, command = false) #function returning filename after conversion of variables
	if settings['cd'].various_artists.empty?
	 	filename =  File.join(settings['basedir'], settings['naming_normal'])
	else
		filename = File.join(settings['basedir'], settings['naming_various'])
	end
	filename = File.expand_path(filename) #Replace any ~/ with the relevant information, such as /home/username/
	
	if command != false
		command.gsub!('%o', filename) #replace the %o with the output filename
		filename = command
	end
	
	if (filename.include?('%va') && settings['cd'].various_artists.empty?)
		filename.gsub!('%va', '')
		puts "Warning: you are using the various artist (%va) in the naming scheme for normal cd's!"
		puts "This is automatically removed"
	end
	{'%a' => settings['cd'].artist, '%b' => settings['cd'].album, '%f' => codec, '%g' => settings['cd'].genre, '%y' => settings['cd'].year, '%n' => sprintf("%02d", track), '%va' => settings['cd'].various_artists[track - 1], '%t' => settings['cd'].tracklist[track - 1]}.each{|key,value|  if filename.include?(key) : filename.gsub!(key, value) end}
	return filename
end

class Gui_support
attr_reader :update_ripping_progress, :update_encoding_progress, :append_2_log, :summary, :ripping_progress, :encoding_progress, :mismatch, :short_summary
	include GetText
	bindtextdomain("rr_lib")

	def initialize(gui, dirnames) #gui is an instance of the graphical user interface used
		@gui = gui
		@logfiles = Array.new
		dirnames.each{|dir| @logfiles << File.open(dir + "/ripping.log", "a")} # Open a log file in append mode for each directory
		@problem_tracks = Hash.new # key = tracknumber, value = new dictionary with key = seconds_chunk, value = [amount_of_chunks, trials_needed]
		@not_corrected_tracks = Array.new # Array of tracks that weren't corrected within the maximum amount of trials set by the user
		@ripping_progress = 0.0
		@encoding_progress = 0.0
		@short_summary = ""
		update_logfiles(_("This log is created by Rubyripper, version %{version}\n") % {:version=>$rr_version})
		update_logfiles(_("Website: http://code.google.com/p/rubyripper\n\n"))
	end
	
	def update_ripping_progress(new_value, calling_function = false) #new_value = float, 1 = 100%
		new_value <= 1.0 ? @ripping_progress = new_value : @ripping_progress = 1.0
		@gui.update("ripping_progress", @ripping_progress)
	end
	
	def update_encoding_progress(new_value, calling_function = false) #new_value = float, 1 = 100%
		new_value <= 1.0 ? @encoding_progress = new_value : @encoding_progress = 1.0
		@gui.update("encoding_progress", @encoding_progress)
	end
	
	def append_2_log(message, calling_function = false)
		@logfiles.each{|logfile| logfile.print(message); logfile.flush()} # Append the messages to the logfiles
		@gui.update("log_change", message)
	end
	
	def update_logfiles(message, summary = false)
		@logfiles.each{|logfile| logfile.print(message); logfile.flush()} # Append the messages to the logfiles
		if summary : @short_summary += message end
	end
	
	def mismatch(track, trial, indexes_with_errors, size, length)
		if !@problem_tracks.key?(track) #First time we encounter this track (Secure_rip->analyze_files() )
			@problem_tracks[track] = Hash.new # create the Hash for the track
			indexes_with_errors.each do |index_of_chunk|
				seconds = index_of_chunk / 176400 # position of chunk rounded in seconds, each second = 176400 bytes
				if !@problem_tracks[track].key?(seconds)
					@problem_tracks[track][seconds] = [1, trial] # different_chunks, trial. First time we encounter this position, so different_chunks = 1
				else
					@problem_tracks[track][seconds][0] += 1 # one more chunk at the same second
				end
			end
		else
			indexes_with_errors.each do |index_of_chunk|
				seconds = index_of_chunk / 176400 # position of chunk rounded in seconds, each second = 176400 bytes
				@problem_tracks[track][seconds][1] = trial #Update the amount of trials needed
			end
		end
		if trial == 0: @not_corrected_tracks << track end #Reached maxtries and still got errors
	end
	
	def summary(matches_all, matches_errors, maxtries) #Give an overview of errors
		update_logfiles(_("\nRIPPING SUMMARY\n\n"), true)
		
		update_logfiles(_("All chunks were tried to match at least %{matches_all}  times.\n") % {:matches_all => matches_all}, true)
		if matches_all != matches_errors: update_logfiles(_("Chunks that differed after %{matches_all} trials,\nwere tried to match %{matches_errors} times.\n") %{:matches_all=>matches_all, :matches_errors=>matches_errors}, true) end

		if @problem_tracks.empty?
 			update_logfiles(_("None of the tracks gave any problems\n"), true)
 		elsif @not_corrected_tracks.size != 0
 			update_logfiles(_("Some track(s) could NOT be corrected within the maximum amount of trials\n"), true)
 			@not_corrected_tracks.each do |track|
				update_logfiles(_("Track %{track} could NOT be corrected completely\n") %{:track=>track}, true)
			end
 		else
 			update_logfiles(_("Some track(s) needed correction,but could\nbe corrected within the maximum amount of trials\n"), true)
 		end

 		if !@problem_tracks.empty? # At least some correction was necessary
			position_analyse(matches_errors, maxtries)
			@short_summary += _("The exact positions of the suspicious chunks\ncan be found in the ripping log\n")
		end
		@logfiles.each{|logfile| logfile.close} #close all the files
 	end
 		
 	def position_analyse(matches_errors, maxtries) # Give an overview of suspicion position in the logfile
		update_logfiles(_("\nSUSPICION POSITION ANALYSIS\n\n"))
		update_logfiles(_("Since there are more than 150 chunks per second, after making the notion of the\n"))
		update_logfiles(_("suspicion position, the amount of initially mismatched chunks for that position is shown.\n\n"))
		@problem_tracks.keys.sort.each do |track| # For each track show the position of the files, how many chunks of that position and amount of trials needed to solve
			update_logfiles(_("TRACK %{track}\n") %{:track=>track})
			@problem_tracks[track].keys.sort.each do |length| #length = total seconds of suspicious position
				minutes = length / 60 # ruby math -> 70 / 60 = 1 (how many times does 60 fit in 70)
				seconds = length % 60 # ruby math -> 70 % 60 = 10 (leftover)
				if @problem_tracks[track][length][1] != 0
					update_logfiles(_("\tSuspicion position : %{minutes}:%{seconds} (%{amount}x) (CORRECTED at trial %{trial}\n") % {:minutes=>sprintf("%02d", minutes), :seconds=>sprintf("%02d", seconds), :amount=>@problem_tracks[track][length][0], :trial=>@problem_tracks[track][length][1] + 1})
				else # Position could not be corrected
					update_logfiles(_("\tSuspicion position : %{minutes}:%{seconds} (%{amount}x) (COULD NOT BE CORRECTED)\n") %{:minutes => sprintf("%02d", minutes), :seconds=>sprintf("%02d", seconds), :amount=>@problem_tracks[track][length][0]})
				end
			end
		end
	end
end

class Cuesheet # Started the very beginning
	def initialize(cddb, output_dirs, file)
		@cddb = cddb
		@output_dirs = output_dirs
		@cuesheet = ''
		@file = file
		toc = `cdparanoia -d #{@cdrom} 2>&1`.split("\n") #command
		
	end
	
	def analyze_toc # fill startpos, endpos, pregap
		@startpos = Array.new #Where does each track start?
		@endpos = Array.new #Where does each track end?
		@pregap = Array.new #Any space between tracks?
		
		while toc[0][0,2] != "==" : toc.shift end #remove all the non-interesting stuff
		toc.shift() #remove the ============ line
			
		@cddb.audiotracks.times do |time|
			temp = toc[time].split() # temp = [track, lengt_sector, length_time, start_sector]
			@startpos << temp[3].to_i # get the startpos
			@endpos << temp[1].to_i + temp[3].to_i #endpos = length + startpos
		end
		
		@startpos.size.times do |time| # Save pregap
			if time == 0: @pregap << @startpos[time]
			else @pregap << @startpos[time] - @endpos[time - 1]
			end
		end
	end
	
	def time(sector) # minutes:seconds:leftover frames
		minutes = sector / 4500 # 75 frames/second * 60 seconds/minute
		seconds = (sector % 4500) / 75
		frames = sector % 75 # leftover
		return "#{sprintf("%02d", minutes)}:#{sprintf("%02d", seconds)}:#{sprintf("%02d", frames)}"
	end
	
	def cuesheet_singlefile()
		@cuesheet += "REM GENRE #{@cddb.genre}\n"
		@cuesheet += "REM DATE #{@cddb.year}\n"
		@cuesheet += "COMMENT \"Rubyripper 0.3\"\n"
		@cuesheet += "PERFORMER \"#{@cddb.artist}\"\n"
		@cuesheet += "TITLE \"#{@cddb.album}\"\n"
		@cuesheet += "FILE \"#{@file}\" WAVE\n" #@file = filename
		@cddb.audiotracks.times do |track|
			@cuesheet += "\tTRACK #{sprintf("%02d", track + 1)} AUDIO\n"
			@cuesheet += "\t\tTITLE \"#{@cddb.tracklist[track]}\"\n"
			@cuesheet += "\t\tPERFORMER \"#{@cddb.artist}\"\n"
			if track > 0 && @pregap[track] > 0: @cuesheet += "\t\tINDEX 00 #{time(@endpos[track - 1])}\n" #Only if there is a pregap
			elsif track == 0 && @pregap[0] > 0: @cuesheet += "\t\tINDEX 00 00:00:00\n"
			end
			@cuesheet += "\t\tINDEX 01 #{time(@startpos[track])}\n"
		end
	end
end

class Cddb

attr_reader :audiotracks, :raw_response, :tracks_seconds, :make_internet_contact, :cdrom, :devicename, :totalsectors, :startsectors, :length_in_sectors, :tracks_playtime, :cd_playtime, :freedb, :freedb_choice, :default_drive, :update_yaml
attr_accessor :artist, :album, :genre, :year, :tracklist, :various_artists

	include GetText
	bindtextdomain("rr_lib")
	
	def initialize(cdrom, instance=false, verbose=false)
		require 'yaml'
		@cdrom = cdrom
		@gui = instance
		@verbose = verbose
		@various_artists = Array.new
		
		if installed('cdparanoia')
			scan_drive() #save the disc info generated by cdparanoia into @output
			parse_output() #get usefull variables from @output
			set_default_values() # default metadata
		else
			@gui.update("error", _("cdparanoia not found. Can't analyze CD!"))
		end
	end
	
	def scan_drive
		@default_drive = false # will be true if cdparanoia doesn't support the device parameter (-d)
		@output = `cdparanoia -d #{@cdrom} -vQ 2>&1` # fire a scan of the disc
		if @output.include?('USAGE')
			if @verbose : puts _("No disc found or cdparanoia doesn't support the drive parameter. Will retry with default drive") end
			@output = `cdparanoia -vQ 2>&1`
			if @output.include?("Unable to open disc") #No disc found by cdparanoia
				@gui.update("error", _("No audio disc found in %{cdrom}") % {:cdrom=>@cdrom}) 
			else 
				@default_drive = true # Apparently the -d parameter isn't supported. Value of default_drive will be used by Secure_RIp class
			end
		end
	end

	def parse_output # based on cdparanoia output
		@audiotracks = 0 ; 	@totalsectors = 0 ; @data_start = 0 ; @data_end = 0 #initialize values to 0, so we can later check if they're != 0
		@startsectors = Array.new ; @length_in_sectors = Array.new ; @tracks_playtime = Array.new ; @tracks_seconds = Array.new # arrays with values for each track

		@devicename = _("Unknown") ; @cd_playtime = '00:00'
		
		@output.split("\n").each do |line|
			if line[0,5] =~ /\s+\d+\./ #Bingo! We've got the info of a track s = space, d = digit, + means 1 or more
				trackinfo = line.split() # Split the values tracknumber, length in sectors, playtime, startsector, starttime
				@audiotracks += 1 # each line present one track
				@startsectors << trackinfo[3].to_i #save the integer value
				@length_in_sectors << trackinfo[1].to_i
				@tracks_playtime << trackinfo[2][1,5]
			elsif line =~ /CDROM\D*:/
				@devicename = $'.strip()
			elsif line[0,5] == "TOTAL"
				@totalsectors = line.split()[1].to_i
				@cd_playtime = line.split()[2][1,5]
			elsif line[0,15] == "track_num = 176"   # 176 = 0xb0 = end of data track??? # only MacOS version prints this
				@data_end = line.split()[5].to_i
			end
		end

		@length_in_sectors.each{|sectors| @tracks_seconds << sectors / 75} # Each second has 75 sectors
		if @data_end != 0   # detected a data track; assume last track == data track
			rawtoc = @output.find_all{|line| line[0,9] == "track_num"}
			@data_start = rawtoc.pop().split()[5].to_i
		end
	end
	
	def set_default_values # for file naming and metadata
		@artist = _('Unknown') ; @album = _('Unknown') ; @genre = _('Unknown'); @year = '0'
		@tracklist = Array.new # Titles of the tracks
		@audiotracks.times{|number| @tracklist << _("Track %{number}") % {:number=> number + 1}}# temporary tracklist
	end

	def gen_freedb_string # manual generate freedb strings! Thanks Scott for providing the code :-)
		checksum = 0
		ssectors = @startsectors
		ntracks = @audiotracks
		if @data_end == 0 # not a data disc (?) so behave normally
			tsectors = @totalsectors
		else
			ntracks += 1
			tsectors = @data_end
			ssectors << @data_start
		end

		ntracks_sectors = ntracks.to_s + " "
		ssectors.each do |sector|
			ntracks_sectors += (sector + 150).to_s + " " #append start sectors, adding mysterious 150 frame offset
			second = (sector + 150) / 75 #convert sectors -> seconds
			second.to_s.each_byte{|digit| checksum += digit.chr.to_i }  #sum digits of track _second_ offset
		end
		checksum %= 255
		hex_id =  checksum.to_s(16).rjust(2, '0') #cksum to hex string
		hex_id += (tsectors / 75).to_s(16).rjust(4, '0') #total playtime to hex string
                hex_id += ntracks.to_i.to_s(16).rjust(2, '0') #track count to hex string

		# freedb_string = hex id + number of tracks + each track sector + total sectors
		@freedb_string = hex_id + " " + ntracks_sectors + ((tsectors + 150) / 75).to_s
        end
	
	def freedb(freedb_options=false, choose=false) # based on cd-discid output
		@freedb_options = freedb_options ? freedb_options : {"site" => "freedb.org", "username" => "anonymous", "hostname" => "my_secret.com"} # if not set, use defaults
		@choose = choose # pass multiple freedb hits to gui

		if installed('discid') # check if we can use discid program
			if RUBY_PLATFORM.include?('darwin') : `diskutil unmount #{@cdrom}` end
			@freedb_string = `discid #{@cdrom}`
			if RUBY_PLATFORM.include?('darwin') : `diskutil mount #{@cdrom}` end
		elsif installed('cd-discid') # if no discid exists, try cd-discid
			@freedb_string = `cd-discid #{@cdrom}`
		else # if both do not exist generate it ourselves (less foolproof)
			@gui.update("error", _("discid or cd-discid not found. Trying fallback..."))
			gen_freedb_string()
			puts _("Using fallback... freedb string = %{freedb_string}" %{:freedb_string=>@freedb_string})
		end
		get_disc_info()
	end
		
	def get_disc_info
 		if @freedb_string != "" # audio-disc is found
			if File.exists?(yaml_file = File.join(ENV['HOME'], '.rubyripper/freedb.yaml')) # Perhaps loaded the disc before?
				@freedb_hash = YAML.load(File.open(yaml_file))
			else
				@freedb_hash = Hash.new
			end
			
			if @freedb_hash.has_key?(@freedb_string) # do we have the freedb string in our collection of saved queries?
				@raw_response = @freedb_hash[@freedb_string]
				@tracklist.clear() #Now fill it with the real tracknames
				cd_vars()
				@gui.update("cddb_hit", false) # Give the signal that we're finished
			else
				make_internet_contact()
			end	
 		else
 			@gui.update("error", _("No audio disc found in %{cdrom}") % {:cdrom=>@cdrom})
 		end
	end
	
	def make_internet_contact
		require 'socket'
		require 'timeout'
		begin
			Timeout::timeout(10) do |timeout_length| #give 10 seconds to get contact
				@socket = TCPSocket.new(@freedb_options['site'], 8880)
				@clientname = 'rubyripper'
				@version = '0.4'
			end
			handshake()
			cddb_query()
		rescue Timeout::Error
			@gui.update("error", _("Couldn't fetch the freedb info within 10 seconds.\n\nDefault values will be shown..."))
			sleep(3) #give some time to read the message
			@gui.update("cddb_hit", false) # Give the signal that we're finished
		rescue
			@gui.update("error", _("Couldn't connect to freedb server. Network down?\n\nDefault values will be shown..."))
			sleep(3) #give some time to read the message
			@gui.update("cddb_hit", false) # Give the signal that we're finished
		end
	end	
		
	def handshake # make first contact and set protocol to 6
		@socket.gets # Get general welcome message
		@socket.puts("cddb hello #{@freedb_options["username"]} #{@freedb_options["hostname"]} #{@clientname} #{@version}") #Identify ourselves
		@socket.gets # Get specific welcome message
		@socket.puts('proto 6') # Set protocol to the latest & greatest
		@socket.gets # Get confirmation
	end
	
	def cddb_query # ask for matches on cd, if there are multiple, interaction with user is possible
		@socket.puts "cddb query #{@freedb_string}" # Fire query
		answer = @socket.gets # The first 3 numbers in the answer indicates if none, one or multiple hits are found
		if answer[0..2] == '200'  #There was only one hit found
			temp, @category, @discid = answer.split()
			freedb_choice(0, true)
		elsif answer[0..2] == '211' || answer[0..2] == '210' #Multiple hits were found
			get_choices()
			if (!@choose || @choices.length < 3) : freedb_choice(0) #Always choose the first one
			else @gui.update("cddb_hit", @choices) #Let the user choose
			end	
		elsif answer[0..2] == '202'
			@gui.update("error", _("No match in Freedb database. Default values are used."))
			@gui.update("cddb_hit", false)
		else
			@gui.update("error", _("cddb_query return code = %{return_code}. Return code not supported.") % {:return_code=>answer[0..2]})
			@gui.update("cddb_hit", false)
		end
	end
	
	def freedb_choice(choice, single=false) #if single=true there's no choise, otherwhise the choice is used for raw_response
		unless single # unless always first choice is used
			if choice == (@choices.length - 1) # keep defaults, don't get freedb (see get_choices function)
				@gui.update("cddb_hit", false)
				@socket.close
				return true
			end
			@category, @discid = @choices[choice].split
		end
		get_raw_response()
		@tracklist.clear() #Now fill it with the real tracknames
		cd_vars()
		@gui.update("cddb_hit", false) # Give the signal that we're finished
	end

	def get_choices	#Puts the multiple choices in an array
		@choices = Array.new
		until (answer = @socket.gets.strip()) == '.' # This means end of output
			@choices << answer
		end
		@choices << _("Keep defaults / don't use freedb") #also use the option to keep defaults
	end

	def get_raw_response #Retrieve all usefull metadata into @raw_response
		@socket.puts("cddb read #{@category} #{@discid}") # Get match
		answer = @socket.gets # Announcement of match
		@raw_response = Array.new
		until (answer = @socket.gets.strip()) == '.' # This means end of output
			if answer != nil && answer[0,1] != '#' && answer[-1,1] != '=' # Skip the garbage
 				@raw_response << answer
			end
		end
		@socket.puts("quit") # Say goodbye to server
		@socket.gets # Confirmation
		@socket.close
		save_response()
	end
	
	def save_response(response = false) #now save the @raw_response(false) or the modified metadata (response=true)
		if not File.directory?(dirname = File.join(ENV['HOME'], '.rubyripper')) #check if dir exists
			Dir.mkdir(dirname)
		end

		if File.exists?(yaml_file = File.join(ENV['HOME'], '.rubyripper/freedb.yaml')) #check if file exists, if not then @freedb_hash is an empty hash
			@freedb_hash = YAML.load(File.open(yaml_file))
		end

		if response != false : @freedb_hash[@freedb_string] = response else @freedb_hash[@freedb_string] = @raw_response end
		yaml_file = File.new(yaml_file, 'w')
		yaml_file.write(@freedb_hash.to_yaml)
		yaml_file.close()
	end
	
	def update_yaml
		raw_response = String.new #now build a Freedb string
		raw_response += "DTITLE=#{@artist} \/ #{@album}\n"
		raw_response += "DYEAR=#{@year}\n"
		raw_response += "\nDGENRE=#{@genre}\n"
		@tracklist.each_index{|index| raw_response += "TTITLE#{index}=#{@tracklist[index]}\n"}
		save_response(raw_response)
		clean_vars() #get any weird characters out of the filenames
	end
	
	def cd_vars #Make some usefull variables from the raw_response.
		@raw_response.each do |line|
			line.strip! #remove any newline characters
			if line =~ /DTITLE=/
				if @artist == _('Unknown') #first time we look at a DTITLE field (can have two lines at maximum)
					(@artist, @album) = $'.split(/\s*\/\s*/) # Remove the '/' with spaces around it, example DTITLE= Judas Priest     /     Sin After Sin
					if @artist == nil: @artist = _('Unknown') end
					if @album == nil: @album = _('Unknown') end
				elsif $' != nil # 2nd line with DTITLE, assume the second line is the continuation of the album name
					buf = $'
					@album = "#{@album}#{buf}"
				end
			elsif line =~ /DYEAR=/
				@year = $' ; if @year == nil : @year = 0 else @year end
			elsif line =~ /DGENRE=/
				@genre = $' ; if @genre == nil : @genre = _('Unknown') end
			elsif line =~ /TTITLE\d*=/
				trackname = $' # may also include variable artist
				line =~ /\d+/ ; tracknumber = $&.to_i #ruby magic, $& == the just matched regular expression
				if trackname != nil && @tracklist.empty? #1st track
					@tracklist << trackname
				elsif trackname != nil && @tracklist.length == tracknumber #counting of tracknumber starts with 0
					@tracklist << trackname 
				elsif trackname != nil #already had a line for this track, so add this trackname to the previous one
					@tracklist[-1] += trackname
				end
			end
		end
		check_for_various()
	end

#Various artist albums have different ways to show the artist and trackname
#Most common notation is ARTIST / TITLE
#Next to that is ARTIST - TITLE
#Then we got "TITLE" by ARTIST

	def check_for_various
		slash_sep = true ; hyphen_sep = true; by_sep = true
		@tracklist.each do |trackname| #walk through all tracks and see if in each track the separator is found
			unless trackname.include?('/')  : slash_sep = false end
			unless trackname.include?('-') : hyphen_sep = false end
			unless trackname.include?('by') : by_sep = false end
		end
		#if a separator is found, split the values accordingly, removing any spaces around the separator
		if slash_sep : @tracklist.each_index{|index| @various_artists[index], @tracklist[index] = @tracklist[index].split(/\s*\/\s*/)}
		elsif hyphen_sep : @tracklist.each_index{|index| @various_artists[index], @tracklist[index] = @tracklist[index].split(/\s*-\s*/)}
		elsif by_sep : @tracklist.each_index{|index| @tracklist[index], @various_artists[index] = @tracklist[index].split(/\s*by\s*/)}
		end
	end
	
	def clean_vars
		@album = clean(@album)
		@artist = clean(@artist)
		@year = clean(@year)
		@genre = clean(@genre)
		@tracklist.each_index{|index| @tracklist[index] = clean(@tracklist[index])}
		@various_artists.each_index{|index| @various_artists[index] = clean(@various_artists[index])}
	end

#The following characters are not allowed in Bash (in combination with other commands): the backquote( ` ), the double quote( " ), ( [ ), ( ] ), the forward slash ( / ), the backward slash ( \ )
#ruby passes by reference, functions with a '!' will immediately save the original value
	def clean(var)
		var.gsub!('`', "'") #replace a backquote with a single quote
		var.gsub!('"', '\"') #Add a slash before the double quote chars, otherwise the shell will complain
		var.gsub!('[', '(') #replace the ' [ ' with a ' ( '
		var.gsub!(']', ')') #replace the ' ] ' with a ' ) '
		var.gsub!('/', ' ') #remove any '/'
		var.gsub!('\\', '') #remove any '\', the \\ means a normal \
		var.gsub!('_', ' ') # replace any underscores with spaces, some freedb info got underscores instead of spaces
		var.gsub!(':', ',') #replace any colons with a comma
		var = var.strip
	end
end

class Secure_rip
	include GetText
	bindtextdomain("rr_lib")

	def initialize(track, settings)
		@track = track
		@settings = settings
		@gui = @settings['log']
		@req_matches_all = @settings['req_matches_all'] # Matches needed for all chunks
		@req_matches_errors = @settings['req_matches_errors'] # Matches needed for chunks that didn't match immediately
		if @settings['offset'] != 0 && @track == @settings['cd'].audiotracks && @settings['rippersettings'].include?('-Z') #workaround for bug in cdparanoia
			@settings['rippersettings'].gsub!(/-Z\s?/, '') #replace the -Z setting (and one space if it is there) with nothing. See issue nr. 13.
		end
		main()
#eject the disc when finished if user checked this option
		if @settings['eject'] && @track == @settings['tracks_to_rip'][-1]
			 if installed('eject') : `eject #{@settings['cd'].cdrom}`
			elsif installed('diskutil'): `diskutil eject #{@settings['cd'].cdrom}` #Mac users don't got eject, but diskutil
			else puts _("No eject utility found!")
			end
		end
	end
	
	def main
		@errors = Hash.new()
		@filesizes = Array.new
		@trial = 0
		
		@req_matches_all.times do # The amount of matches required tells how many times the files should be ripped before comparing
			@trial += 1
			rip()
			@filesizes << File.size("#{@settings['temp_dir']}track#{@track}_#{@trial}.wav")
		end
		
		test_for_filesize() #Make sure that we got the needed files with the same size
		analyze_files() #If there are differences, save them in the @errors hash
				
		while @errors.size > 0
			if @trial > @settings['max_tries'] && @settings['max_tries'] != 0 # We would like to respect our users settings, wouldn't we?
				@gui.append_2_log(_("Maximum tries reached. %{amount} chunk(s) didn't match the required %{amount2} times\n") % {:amount=>@errors.length, :amount2=>@req_matches_errors})
				@gui.append_2_log(_("Will continue with the file we've got so far\n"))
				@gui.mismatch(@track, 0, @errors.keys, @size_match, @settings['cd'].tracks_seconds[@track - 1]) # zero means it is never solved.
				break # break out loop and continue using trial1
			end
			
			@trial += 1
			rip()
			
			if File.size("#{@settings['temp_dir']}track#{@track}_#{@trial}.wav") == @size_match # The file size must match with previous attempts
				if @trial > @req_matches_errors : wave_out() else read_positions_only() end # If the req_matches errors is equal of higher to @trial, no match would ever be found, so skip
			else
				@gui.append_2_log(_("File size of trial %{trial} does not match previous attempts\n") %{:trial=> @trial})
			end
		end
		
		get_digest() # Get a MD5-digest for the logfile
		
		@gui.update_ripping_progress(@settings['rip_percentages'].shift)
	end
	
	def test_for_filesize
		@size_match = false # Until proven otherwise :P
		if @filesizes.uniq.size == 1 : @size_match = @filesizes[0] end # When there is no difference in the filesize proceed (this should happen in 99% of all cases)
		while !@size_match
			@gui.append_2_log(_("Still not %{amount} times the same filesize. Trying another time\n") %{:amount=>@req_matches_all})
			@trial += 1 ; rip() ; @filesizes << File.size("#{@settings['temp_dir']}track#{@track}_#{@trial}.wav")
			filesizes = @filesizes.sort() # Make a sorted copy
			filesizes.each do |size| # This look like a lot of iteration, but most of the code is only run once
				if filesizes.rindex(size) - filesizes.index(size) >= (@req_matches_all - 1) # Do we have the required amount of same sizes?
					@size_match = size
					trialnumber = 1 # So we have to iterate again with trialnumber as a help value.
					@filesizes.each_index do |index| # now clean up files with wrong filesizes. Since we sorted the copy, we don't know which trials have wrong filesizes.
						if @filesizes[index] == @size_match 
							if File.exist?("#{@settings['temp_dir']}track#{@track}_#{trialnumber}.wav") : trialnumber += 1 # The file won't exist if it is deleted earlier in the loop
							else File.rename("#{@settings['temp_dir']}track#{@track}_#{index + 1}.wav", "#{@settings['temp_dir']}track#{@track}_#{trialnumber}.wav") ; trialnumber += 1
							end	
  						else
  							File.delete("#{@settings['temp_dir']}track#{@track}_#{index + 1}.wav") # Delete files with non-matching filesizes
  						end
  					end
  					break # Since we have a filesize match, don't finish the loop
				end
			end
		end
	end
	
	def analyze_files()
		@gui.append_2_log(_("Analyzing files for mismatching chunks\n"))
		files = Array.new
		@req_matches_all.times do |time|
			files << File.new("#{@settings['temp_dir']}track#{@track}_#{time + 1}.wav", 'r')
		end
				
		(@req_matches_all - 1).times do |time|
			index = 0 ; files.each{|file| file.pos = 44} # 44 = wav container overhead, 2352 = size for a audiocd sector as used in cdparanoia
			while index + 44 < @size_match
				if !@errors.key?(index) && files[0].read(2352) != files[time + 1].read(2352) # Does this sector matches the previous ones? and isn't the position already known?
					files.each{|file| file.pos = index + 44} # Reset each read position of the files
					@errors[index] = Array.new
					files.each{|file| @errors[index] << file.read(2352)} # Save the chunk for all files in the just created array
				end
				index += 2352
			end
		end
		files.each{|file| file.close}
		(@req_matches_all - 1).times{|time| File.delete("#{@settings['temp_dir']}track#{@track}_#{time + 2}.wav")} # Don't need these files anymore. Differences are saved in memory.
		if @errors.size == 0
			@gui.append_2_log(_("Every chunk matched %{amount} times :)\n") %{:amount=>@req_matches_all})
		else
			@gui.mismatch(@track, @trial, @errors.keys, @size_match, @settings['cd'].tracks_seconds[@track-1]) # report for later position analysis
			@gui.append_2_log(_("%{amount} chunk(s) didn't match %{amount2} times.\n") % {:amount=>@errors.length, :amount2=>@req_matches_all})
		end
	end
	
	def read_positions_only # This is used if the required amount of matches is too big to ever be realized
		file = File.new("#{@settings['temp_dir']}track#{@track}_#{@trial}.wav", 'r')
		@errors.keys.sort.each do |start_chunk|
			file.pos = start_chunk + 44 # 44 = wav container overhead, 2352 = size for a audiocd sector as used in cdparanoia
			@errors[start_chunk] << file.read(2352)
		end
		file.close
		File.delete("#{@settings['temp_dir']}track#{@track}_#{@trial}.wav") # Don't need this file anymore. Differences are saved in memory.
		@gui.mismatch(@track, @trial, @errors.keys, @size_match, @settings['cd'].tracks_seconds[@track-1]) # report for later position analysis
	end
	
	def wave_out
	#Let the errors 'wave' out. Chunks with different outcomes must be matches @req_matches_errors times.
		file1 = File.new("#{@settings['temp_dir']}track#{@track}_1.wav", 'r+')
		file2 = File.new("#{@settings['temp_dir']}track#{@track}_#{@trial}.wav", 'r')
		@errors.keys.sort.each do |start_chunk| # Sort the hash keys to prevent jumping forward and backwards in the file
			file2.pos = start_chunk + 44 # 44 = wav container overhead, 2352 = size for a audiocd sector as used in cdparanoia
			@errors[start_chunk] << temp = file2.read(2352)
			@errors[start_chunk].sort! # Sort the array before looking if temp solves our requirement of matches
			if (@errors[start_chunk].rindex(temp) - @errors[start_chunk].index(temp)) == (@req_matches_errors - 1) # Matches are r(ight)index - index because array is sorted
				file1.pos = start_chunk
				file1.write(temp)
				@errors.delete(start_chunk)
			end
		end
		file1.close
		file2.close
		File.delete("#{@settings['temp_dir']}track#{@track}_#{@trial}.wav") # Don't need this file anymore. Differences are saved in memory.
		
		if @errors.size == 0
			@gui.append_2_log(_("Error(s) succesfully corrected, %{amount} matches found for each chunk :)\n") % {:amount=>@req_matches_errors})
		else
			@gui.mismatch(@track, @trial, @errors.keys, @size_match, @settings['cd'].tracks_seconds[@track-1]) # report for later position analysis
			@gui.append_2_log(_("%{amount} chunk(s) didn't match %{amount2} times.\n") % {:amount=>@errors.length, :amount2=>@req_matches_errors})
		end
	end
	
	def rip # set cdparanoia command + parameters
		@gui.append_2_log(_("Starting to rip track %{track}, trial %{trial}\#\n") %{:track=>@track, :trial=>@trial})
		command = "cdparanoia"
		if @settings['rippersettings'].size != 0 : command += " #{@settings['rippersettings']}" end 
		command += (" #{@track}") # second parameter tracknumber
		unless @settings['cd'].default_drive : command += " -d #{@settings['cdrom']}" end # the ported cdparanoia for MacOS misses the -d option, default drive will be used.
		command += " -O #{@settings['offset']}"
		command += " \"#{@settings['temp_dir']}track#{@track}_#{@trial}.wav\"" #output name
		unless @settings['verbose'] : command += " 2>&1" end # hide the output of cdparanoia output
		`#{command}` #Launch the cdparanoia command
	end
	
	def get_digest
		digest = Digest::MD5.new()
		file = File.open("#{@settings['temp_dir']}track#{@track}_1.wav", 'r')
		index = 0
		while (index < @size_match)
			digest << file.read(100000)
			index += 100000
		end
		file.close()
		@gui.append_2_log(_("MD5 sum: %{md5sum}\n\n") %{:md5sum=>digest.hexdigest})
	end
end

class Encode
	def initialize(track, settings)
		@track = track
		@settings = settings
		@artist = @settings['cd'].various_artists[@track - 1]
		@gui = @settings['log']
		@trackname = @settings['cd'].tracklist[@track - 1]
		percentages()
		ENV['CHARSET'] = "UTF-8" #Needed for correct character handling for the oggenc executable. Perhaps other encoders use it as well.
		if @settings['flac'] : do_flac() end
		if @settings['vorbis'] : do_vorbis() end
		if @settings['mp3'] : do_mp3() end
		if @settings['wav'] : do_wav() end
		if @settings['other'] : do_other end
		File.delete("#{@settings['temp_dir']}track#{@track}_1.wav") #We don't need the wav file after it's encoded.
	end
	
	def percentages
		if @settings['tracks_to_rip'][0] == @track : @gui.update_encoding_progress(0.0) end # set it to 0% for the first track, so the gui shows it's started.
		@codecs = 0
		if @settings['flac']: @codecs += 1 end
		if @settings['vorbis']: @codecs += 1 end
		if @settings['mp3']: @codecs += 1 end
		if @settings['wav']: @codecs += 1 end
		if @settings['other']: @codecs += 1 end
		@complete = @settings['enc_percentages'].shift()
	end

	def do_flac
		filename = get_filename(@settings,  'flac', @track) + '.flac'
		if !@settings['flacsettings'] : @settings['flacsettings'] = '--best' end
		flac(filename)
		update_playlist('flac', filename) if @settings['playlist']
		increase = (@complete - @gui.encoding_progress) / @codecs
		@gui.update_encoding_progress(@gui.encoding_progress + increase)
		@codecs -= 1
	end
		
	def do_vorbis
		filename = get_filename(@settings, 'vorbis', @track) + '.ogg'
		if !@settings['vorbissettings'] : @settings['vorbissettings'] = '-q 6' end
		vorbis(filename)
		update_playlist('vorbis', filename) if @settings['playlist']
		increase = (@complete - @gui.encoding_progress) / @codecs
		@gui.update_encoding_progress(@gui.encoding_progress + increase)
		@codecs -= 1
	end
		
	def do_mp3
		@possible_lame_tags = ['A CAPPELLA', 'ACID', 'ACID JAZZ', 'ACID PUNK', 'ACOUSTIC', 'ALTERNATIVE', 'ALT. ROCK', 'AMBIENT', 'ANIME', 'AVANTGARDE', \
'BALLAD', 'BASS', 'BEAT', 'BEBOB', 'BIG BAND', 'BLACK METAL', 'BLUEGRASS', 'BLUES', 'BOOTY BASS', 'BRITPOP', 'CABARET', 'CELTIC', 'CHAMBER MUSIC', 'CHANSON', \
'CHORUS', 'CHRISTIAN GANGSTA RAP', 'CHRISTIAN RAP', 'CHRISTIAN ROCK', 'CLASSICAL', 'CLASSIC ROCK', 'CLUB', 'CLUB-HOUSE', 'COMEDY', 'CONTEMPORARY CHRISTIAN', \
'COUNTRY', 'CROSSOVER', 'CULT', 'DANCE', 'DANCE HALL', 'DARKWAVE', 'DEATH METAL', 'DISCO', 'DREAM', 'DRUM & BASS', 'DRUM SOLO', 'DUET', 'EASY LISTENING', \
'ELECTRONIC', 'ETHNIC', 'EURODANCE', 'EURO-HOUSE', 'EURO-TECHNO', 'FAST-FUSION', 'FOLK', 'FOLKLORE', 'FOLK/ROCK', 'FREESTYLE', 'FUNK', 'FUSION', 'GAME', \
'GANGSTA RAP', 'GOA', 'GOSPEL', 'GOTHIC', 'GOTHIC ROCK', 'GRUNGE', 'HARDCORE', 'HARD ROCK', 'HEAVY METAL', 'HIP-HOP', 'HOUSE', 'HUMOUR', 'INDIE', 'INDUSTRIAL', \
'INSTRUMENTAL', 'INSTRUMENTAL POP', 'INSTRUMENTAL ROCK', 'JAZZ', 'JAZZ+FUNK', 'JPOP', 'JUNGLE', 'LATIN', 'LO-FI', 'MEDITATIVE', 'MERENGUE', 'METAL', 'MUSICAL', \
'NATIONAL FOLK', 'NATIVE AMERICAN', 'NEGERPUNK', 'NEW AGE', 'NEW WAVE', 'NOISE', 'OLDIES', 'OPERA', 'OTHER', 'POLKA', 'POLSK PUNK', 'POP', 'POP-FOLK', 'POP/FUNK', \
'PORN GROOVE', 'POWER BALLAD', 'PRANKS', 'PRIMUS', 'PROGRESSIVE ROCK', 'PSYCHEDELIC', 'PSYCHEDELIC ROCK', 'PUNK', 'PUNK ROCK', 'RAP', 'RAVE', 'R&B', 'REGGAE', \
'RETRO', 'REVIVAL', 'RHYTHMIC SOUL', 'ROCK', 'ROCK & ROLL', 'SALSA', 'SAMBA', 'SATIRE', 'SHOWTUNES', 'SKA', 'SLOW JAM', 'SLOW ROCK', 'SONATA', 'SOUL', 'SOUND CLIP', \
'SOUNDTRACK', 'SOUTHERN ROCK', 'SPACE', 'SPEECH', 'SWING', 'SYMPHONIC ROCK', 'SYMPHONY', 'SYNTHPOP', 'TANGO', 'TECHNO', 'TECHNO-INDUSTRIAL', 'TERROR', 'THRASH METAL', \
'TOP 40', 'TRAILER', 'TRANCE', 'TRIBAL', 'TRIP-HOP', 'VOCAL']
		filename = get_filename(@settings, 'mp3', @track) + '.mp3'
		if !@settings['mp3settings'] : @settings['mp3settings'] = "--preset fast standard" end
		if @possible_lame_tags.include?(@settings['cd'].genre.upcase): genre = @settings['cd'].genre else genre = 'other' end
		mp3(filename, genre)
		update_playlist('mp3', filename) if @settings['playlist']
		increase = (@complete - @gui.encoding_progress) / @codecs 
		@gui.update_encoding_progress(@gui.encoding_progress + increase)
		@codecs -= 1
	end
		
	def do_wav	
		filename = get_filename(@settings, 'wav', @track) + '.wav'
		wav(filename)
		update_playlist('wav', filename) if @settings['playlist']
		increase = (@complete - @gui.encoding_progress) / @codecs 
		@gui.update_encoding_progress(@gui.encoding_progress + increase)
		@codecs -= 1
	end
	
	def do_other
		command = get_filename(@settings, 'other', @track, @settings['othersettings'].dup) #pass the commandline for other and replace all % fields (except %i)
		command.gsub!('%i', " \"#{@settings['temp_dir']}track#{@track}_1.wav\"") # %i = input filename
		
		if @verbose : puts command end
 		`#{command}` # execute command
		
 		update_playlist(extension, filename) if @settings['playlist']
		increase = (@complete - @gui.encoding_progress) / @codecs
		@gui.update_encoding_progress(@gui.encoding_progress + increase)
		@codecs -= 1
	end
	
	def flac(filename)
		command = %Q{flac #{@settings['flacsettings']} -o "#{filename}" --tag ARTIST="#{@artist ? @artist : @settings['cd'].artist}" --tag TITLE="#{@trackname}" --tag ALBUM="#{@settings['cd'].album}" --tag DATE="#{@settings['cd'].year}" --tag TRACKNUMBER="#{@track}" --tag GENRE="#{@settings['cd'].genre}" "#{@settings['temp_dir']}track#{@track}_1.wav"#{" 2>&1" unless @settings['verbose']}}
		`#{command}` #execute command
	end
	
	def vorbis(filename)
		command = %Q{oggenc -o "#{filename}" #{@settings['vorbissettings']} -c DATE="#{@settings['cd'].year}" -c TRACKNUMBER="#{@track}" -c TITLE="#{@trackname}" -c ALBUM="#{@settings['cd'].album}" -c ARTIST="#{@artist ? @artist : @settings['cd'].artist}" -c GENRE="#{@settings['cd'].genre}" "#{@settings['temp_dir']}track#{@track}_1.wav" #{" 2>&1" unless @settings['verbose']}}
		`#{command}` #execute command
	end
	
	def mp3(filename, genre)
		command = %Q{lame --id3v2-only #{@settings['mp3settings']} --tt "#{@trackname}" --ta "#{@artist ? @artist : @settings['cd'].artist}" --tl "#{@settings['cd'].album}" --ty "#{@settings['cd'].year}" --tn "#{@track}" --tg "#{genre}" "%i" "%o" #{" 2>&1" unless @settings['verbose']}}
			
		require 'iconv'
		command = Iconv.conv("ISO-8859-1", "UTF-8", command) #translate the UTF-8 string to latin. This is needed because of a lame bug.
		command.sub!('%o', filename) #the %o output file should stay in UTF-8
		command.sub!('%i', "#{@settings['temp_dir']}track#{@track}_1.wav") #the %i input file should stay in UTF-8 as well
		`#{command}` #execute command
	end
	
	def wav(filename)
		require 'fileutils'
		FileUtils.cp("#{@settings['temp_dir']}track#{@track}_1.wav", filename)
	end
	
	def update_playlist(format, filename)
		m3ufile = File.new(File.dirname(filename) + "/#{@settings['cd'].artist} - #{@settings['cd'].album} (#{format}).m3u", 'a')
		m3ufile.puts "#{File.basename(filename)}"
		m3ufile.close()
	end
end

class Rubyripper
attr_reader :settings_ok, :start_rip, :output_dirs, :gui,  :postfix_dir, :overwrite_dir
	include GetText
	bindtextdomain("rr_lib")
	
	def initialize(settings)
		@settings = settings.dup
		@directory = false
		@gui = false
	end
	
	def settings_ok
		if check_configuration && test_for_deps && prepare_dirs
			return true
		else
			return false
		end
	end
	
	def start_rip
		update_gui() # Give some info about the cdrom-player, the codecs, the ripper, cddb_info
		@settings['cd'].update_yaml() #save all the (changed?) metadata on the local disk
		compute_percentage() # Do some pre-work to get the progress updater working later on
		temp_dir() # Check if the temporary dir exists, if not: make it if basedir does exist or fallback to /tmp/rr/.
		require 'digest/md5' # Needed for secure class, only have to load them ones here.
		ripper() # Function to control Secure_rip and Encode class
		cleanup() # Delete any temporary files
		@gui.summary(@settings['req_matches_all'], @settings['req_matches_errors'], @settings['max_tries'])
	end
	
	def check_configuration
		unless File.symlink?(@settings['cdrom']) || File.blockdev?(@settings['cdrom']) # does the cdrom device or a symlink to it exist?
			@settings['instance'].update("error", _("The device %{cdrom} doesn't exist on your system!") %{:cdrom=>@settings['cdrom']})
			return false
		end

		if @settings['tracks_to_rip'].size == 0
			@settings['instance'].update("error", _("Please select at least one track."))
			return false
		end
		
		temp = Cddb.new(@settings['cdrom'], @settings['instance'])
		if @settings['cd'].cd_playtime != temp.cd_playtime # Test if the user has put in another CD to fool me ;)
			@settings['instance'].update("error", _("The Gui doesn't match inserted cd. Please press Scan Drive first."))
 			return false
		end
		
		unless @settings['flac'] || @settings['vorbis'] || @settings['mp3'] || @settings['wav'] || @settings['other'] # Test if at least one codec is selected
			@settings['instance'].update("error", _("No codecs are selected!"))
			return false
 		end
 		
		#Make sure errors get at least as much corrected as non-errors.
 		if @settings['req_matches_all'] > @settings['req_matches_errors'] : @settings['req_matches_errors'] = @settings['req_matches_all'] end
		return true
	end
	
	def test_for_deps()
		{"ripper" => "cdparanoia", "flac" => "flac", "vorbis" => "oggenc", "mp3" => "lame"}.each do |setting, binary|
			if @settings[setting] && !installed(binary)
				@settings['instance'].update("error", _("%{codec} not found on your system!") %{:codec=>binary.capitalize})
				return false
			end
		end
		return true
	end
	
	def prepare_dirs() # Look if the dirs already exist and create them
		if @settings['cd'].various_artists.empty? # if no various artist cd is found
			destination_dir = File.join(@settings['basedir'], @settings['naming_normal'])
		else
			destination_dir = File.join(@settings['basedir'], @settings['naming_various'])
		end
			
		if destination_dir.include?("/%b") && @settings['cd'].album[0,1] == '.' #If a dir or file starts with an album and the album starts with a dot 
			@settings['cd'].album.sub!(/\.*/, '') #remove those dot(s) from the albumname, otherwise they're hidden files in linux.
		end
		
		require 'ftools' # Adding some extra File functions like makedirs
	
		@output_dirs = Array.new
		['flac', 'vorbis', 'mp3', 'wav', 'other'].each do |codec| #save the dirname in @output_dirs
			if @settings[codec] #for each selected codec, save the dirname
				@output_dirs << File.dirname(get_filename(@settings, codec))
				unless destination_dir.include?('%f') : break end #only run the code ones if no codec format is found, otherwise the dirs are the same
			end
		end

		@output_dirs.each do |dir| #look for each dir if it exists
			 if File.directory?(dir)
			 	@settings['instance'].update("dir_exists")
				return false 
			 end
		end
		
		
		@output_dirs.each do |dir| #Look for the highest existing dir and check if we have writing access there.
			while not File.directory?(dir)
				dir = File.dirname(dir) #chop of the subdir if it's not yet created
			end
			if not File.writable?(dir)
				@settings['instance'].update("error", _("Can't create output directory!\nYou have no writing acces in dir %{dir}") %{:dir=>dir})
				return false
			end
		end
		
		@output_dirs.each{|dir| File.makedirs(dir)} #create the dir
		
		@gui = @settings['log'] = Gui_support.new(@settings['instance'], @output_dirs) # The gui communication class :)
		return true
	end
	
	def postfix_dir #add the first free number as a postfix to the output dir
		postfix = 1
		@output_dirs.each{|dir| while File.directory?(dir + "\##{postfix}") : postfix += 1 end} #increase enough for all directories

		if @settings['cd'].various_artists.empty? #use a trick with File.join to remove any trailing '/'
			@settings['naming_normal'] = File.join(File.dirname(@settings['naming_normal']) + "\##{postfix}", File.basename(@settings['naming_normal']))
		else
			@settings['naming_various'] = File.join(File.dirname(@settings['naming_various']) + "\##{postfix}", File.basename(@settings['naming_various']))
		end
		prepare_dirs() #problem is solved, so continue
	end
	
	def overwrite_dir #overwrite the dir
		@output_dirs.each do |dir|
			if File.directory?(dir)
				Dir.foreach(dir){|file| if File.file?(File.join(dir,file)) : File.delete(File.join(dir, file)) end} #first delete all files in the dir
				Dir.rmdir(dir) #then delete the dir itself
			end
		end
		prepare_dirs() #problem is solved, so continue
	end

	def update_gui
		@gui.append_2_log(_("Cdrom player used to rip:\n%{drivename}\n") %{:drivename=>@settings['cd'].devicename})
		@gui.append_2_log(_("Cdrom offset used: %{offset}\n\n") %{:offset=>@settings['offset']})
		@gui.append_2_log(_("Ripper used: cdparanoia %{settings}\n") % {:settings=> if @settings['rippersettings']: @settings['rippersettings'] else _('default settings') end})
		@gui.append_2_log(_("Matches required for all chunks: %{amount}\n") %{:amount=>@settings['req_matches_all']})
		@gui.append_2_log(_("Matches required for erroneous chunks: %{amount}\n\n") % {:amount=>@settings['req_matches_errors']})
		
		@gui.append_2_log(_("Codec(s) used:\n"))
		if @settings['flac']: @gui.append_2_log(_("-flac \t-> %{settings} (%{version})\n") %{:settings=>@settings['flacsettings'], :version=>`flac --version`.strip}) end
		if @settings['vorbis']: @gui.append_2_log(_("-vorbis\t-> %{settings} (%{version})\n") %{:settings=>@settings['vorbissettings'], :version=>`oggenc --version`.strip}) end
		if @settings['mp3']: @gui.append_2_log(_("-mp3\t-> %{settings}\n(%{version}\n") %{:settings=>@settings['mp3settings'], :version=>`lame --version`.split("\n")[0]}) end
		if @settings['wav']: @gui.append_2_log(_("-wav\n")) end
		if @settings['other'] : @gui.append_2_log(_("-other\t-> %{settings}\n") %{:settings=>@settings['othersettings']}) end
		@gui.append_2_log(_("\nCDDB INFO\n"))
		@gui.append_2_log(_("\nArtist\t= ") + @settings['cd'].artist)
		@gui.append_2_log(_("\nAlbum\t= ") + @settings['cd'].album)
		@gui.append_2_log(_("\nYear\t= ") + @settings['cd'].year)
		@gui.append_2_log(_("\nGenre\t= ") + @settings['cd'].genre)
		@gui.append_2_log(_("\nTracks\t= ") + @settings['cd'].audiotracks.to_s + "\n\n")
		@settings['cd'].audiotracks.times do |track|
			@gui.append_2_log("#{sprintf("%02d", track + 1)} - #{@settings['cd'].tracklist[track]}\n")
		end
		@gui.append_2_log(_("\nSTATUS\n\n"))
	end
	
	def compute_percentage
		@settings['rip_percentages'] = Array.new() #cumulative progress
		total_sectors = 0.0 # It can be that the user doesn't want to rip all tracks, so calculate it
		@settings['tracks_to_rip'].each{|track| total_sectors += @settings['cd'].length_in_sectors[track - 1]}
		
		@settings['tracks_to_rip'].each do |track|
			if track == @settings['tracks_to_rip'][-1] # Ensure last value is 1.0
				@settings['rip_percentages'] << 1.0 
			elsif track == @settings['tracks_to_rip'][0] # Cannot refer to the previous value of rip_percentages, since it has none
				@settings['rip_percentages'] << @settings['cd'].length_in_sectors[track - 1] / total_sectors
			else # progress when track is done: length_track / total + previous amount
				@settings['rip_percentages'] << (@settings['cd'].length_in_sectors[track - 1] / total_sectors) + @settings['rip_percentages'][-1]
			end
 		end
 		@settings['enc_percentages'] = @settings['rip_percentages'].dup #Make a copy for the encoding progress.
 	end
 	
	def temp_dir # use a directory named temp where encoding takes place
		@settings['temp_dir'] = File.dirname(@output_dirs[0]) + '/temp/' #Asking the dirname of a directory returns the directory of one higher level
		unless File.directory?(@settings['temp_dir']) : Dir.mkdir(@settings['temp_dir']) end # If the directory not yet exists, create it
	end
		
	def ripper
		@last_track_ripped = false #necessary because of the threaded encoding
		@tracks_to_encode = Array.new() #necessary because of the threaded encoding
		@gui.update_ripping_progress(0.0, "ripper") # Give a hint to the gui that ripping has started
		@settings['tracks_to_rip'].each do |track|
			@secure_rip = Secure_rip.new(track, @settings)
			@tracks_to_encode << track
			if @settings['threading']
				unless @encoding
					@encoding = Thread.new {encode()} #Let the encoding take place in the background.
					@encoding.priority = -1 #Encoding gets lower priority than ripping
				end
			else
				@encode_instance = Encode.new(track, @settings)
			end
		end
		
		@last_track_ripped = true
		if @settings['threading']: @encoding.join() end #Wait untill tracks are encoded, before cleaning up
	end

	def encode()
		until @last_track_ripped && @tracks_to_encode.length == 0
			while track = @tracks_to_encode.shift
				@encode_instance = Encode.new(track, @settings)
			end
			if @tracks_to_encode.length == 0 && @last_track_ripped == false
				sleep(1)
			end
		end
	end
	
	def cleanup
		if File.directory?(@settings['temp_dir'])
			Dir.foreach(@settings['temp_dir']) do |file|
				if file[0,1] != '.' && File.file?("#{@settings['temp_dir']}#{file}"): File.delete("#{@settings['temp_dir']}#{file}") end
			end
			Dir.rmdir(@settings['temp_dir'])
		end
	end
end
